// addspecies : main file
#define ADDSPECIES_MAIN
#include "addspecies.h"

int main(int argc, char** argv) 
{
    string filename;
    string options;
    clock_t start, finish;
    start = clock();

    if (argc > 1 && (argv[1][0] == '?' || 
        (argv[1][0] == '-' || argv[1][0] == '/') &&
        (argv[1][1] == 'H' || argv[1][1] == 'h' || argv[1][1] == '?'))) 
    {
        string pgm(argv[0]);
        try 
        {
            string::size_type pos = pgm.find_last_of(".");
            pgm.erase(pos);
            pos = pgm.find_last_of("\\/");
            if (pos != string::npos)
                pgm.erase(0, pos+1);
        }
        catch(...) { }
        cerr << "\nUsage (v " << version << "): " << pgm << " [options] "
            << "source target [" << def_config << "]\n"
            << " Options:\n"
            << " -En     E-value threshold, 0=none [" << def_E << "]\n"
            << " -Gn     gene displaying step, 0=none [" << def_genestep << "]\n"
            << " -M      write sorted matrix of intergenic scores [no]\n"
            << " -U      write unsorted matrix of intergenic scores [no]\n"
            << " -Qn     quit after: 0=matrices, 1=pass1, 2=para [" << def_quit << "]\n"
            << " -Of     weight threshold for orthology [" << def_oweight << "]\n"
            << " -Pf     weight threshold for paralogy [" << def_pweight << "]\n"
            << " -Wfile  write added orthologs to separate file [no]\n"
            << endl;
        exit(0);
    }

    for (int n = 1; n < argc; n++) 
    {
        char c = argv[n][0];
        if (c != '-' && c != '/')       // non-option found
        {
            if (global.source.empty())
                global.source.assign(argv[n]);
            else if (global.target.empty())
                global.target.assign(argv[n]);
            else if (global.config == def_config)
                global.config.assign(argv[n]);
            else 
            {
                cerr << "Extra argument in command line: " << argv[n] << endl;
                exit(1);
            }
            continue;
        }
        options.append(argv[n]);
        options.append(" ");
        c = argv[n][1];
        string option(argv[n]+2);
        istringstream iss(option);
        switch (c) 
        {
            case 'E': case 'e':     // E-value threshold
                iss >> global.Elimit;
                global.givenElimit = true;
                break;
            case 'G': case 'g':     // gene no. output step
                iss >> global.genestep;
                break;
            case 'U': case 'u':
                global.unsorted = true;
                // throwdown
            case 'M': case 'm':
                global.outmatrix = true;
                break;
            case 'Q': case 'q':
                iss >> global.quit;
                break;
            case 'O': case 'o':
                iss >> global.oweight;
                global.givenOweight = true;
                break;
            case 'P': case 'p':
                iss >> global.pweight;
                global.givenPweight = true;
                break;
            case 'W': case 'w':
                global.addition_file = option;
                break;
            default:
                cerr << "Invalid command line option: -" << c << endl;
                exit(1);
        }
    }

    if (!ReadConfig()) 
    {
        cerr << "Error in config file " << global.config << endl;
        exit(2);
    }

    // check source and target
    for (int k = 0; k < global.nAnimals; k++) 
    {
        Species& spec = global.Animals[k];
        if (spec.name == global.source) 
            global.fromsp = k;
        else if (spec.name == global.target)
            global.tosp = k;
    }
    if (global.source.empty() || global.fromsp < 0) 
    {
        cerr << "Invalid name of source species: " << global.source << endl;
        exit(3);
    }
    if (global.target.empty() || global.tosp < 0) 
    {
        cerr << "Invalid name of target species: " << global.target << endl;
        exit(3);
    }

    cout << "\nUtility for adding non-Ensembl species (version " << version << ")\n\n"
         << "Total species: " << global.nAnimals << "  Options: " << options << "\n"
         << "Add: " << global.source << "(" << global.fromsp << ") -> "
         << global.target << "(" << global.tosp << ") eval=" << global.Elimit 
         << " orth=" << global.oweight << " para=" << global.pweight << "\n" << endl;

    // Read gene tables for all species (among them source and target)
    for (int sp = 0; sp < global.nAnimals; sp++) 
    {
        filename = global.info_path + global.Animals[sp].name + global.info_ext;
        cout << "Reading gene info file " << filename << " ..." << endl;
        if (!ReadGeneInfo(sp)) 
        {
            cerr << "Gene info file error: " << filename << endl;
            exit(4);
        }
    }
    cout << endl;

    // Read table of orthologs and paralogs for target species
    filename = global.homology_path + global.target + global.homology_ext;
    cout << "Reading orthologs from " << filename << " ..." << setw(9) << " "; 
    if (!ReadShortData(global.tosp)) 
    {
        cerr << "Table of orthologs error: " << filename << endl;
        exit(5);
    }
    cout << endl;

    filename = global.paralogs_path + global.target + global.paralogs_ext;
    cout << "Reading paralogs from " << filename << " ..." << setw(9) << " "; 
    if (!ReadShortData(global.tosp, true)) 
    {
        cerr << "Table of paralogs error: " << filename << endl;
        exit(5);
    }
    cout << endl;

    // Read two matrices of scores for <source, target> pair of species
    filename = global.matrix_path + global.source + global.matrix_separ +
        global.target + global.matrix_ext;
    cout << "Reading score matrix " << filename << " ..." << setw(9) << " ";
    if (!ReadMatrix(global.fromsp, global.tosp)) 
    {
        cerr << "Matrix of scores error: " << filename << endl;
        exit(6);
    }
    cout << endl;

    filename = global.matrix_path + global.target + global.matrix_separ +
        global.source + global.matrix_ext;
    cout << "Reading score matrix " << filename << " ..." << setw(9) << " ";
    if (!ReadMatrix(global.tosp, global.fromsp)) 
    {
        cerr << "Matrix of scores error: " << filename << endl;
        exit(6);
    }
    cout << endl;

    // Read two matrices of self-scores for source and target species
    filename = global.matrix_path + global.source + global.matrix_separ +
        global.source + global.matrix_ext;
    cout << "Reading score matrix " << filename << " ..." << setw(9) << " ";
    if (!ReadMatrix(global.fromsp, global.fromsp)) 
    {
        cerr << "Matrix of scores error: " << filename << endl;
        exit(6);
    }
    cout << endl;

    filename = global.matrix_path + global.target + global.matrix_separ +
        global.target + global.matrix_ext;
    cout << "Reading score matrix " << filename << " ..." << setw(9) << " ";
    if (!ReadMatrix(global.tosp, global.tosp)) 
    {
        cerr << "Matrix of scores error: " << filename << endl;
        exit(6);
    }
    cout << endl;

    // Output final matrices of intergenic scores
    if (global.outmatrix) 
    {
        filename = global.outmatrix_path + global.source + global.outmatrix_separ +
            global.target + global.outmatrix_ext;
        cout << "Writing gene matrix " << filename << " ..." << setw(9) << " ";
        if (!WriteMatrix(global.fromsp, global.tosp)) 
        {
            cerr << "Matrix output error: " << filename << endl;
            exit(7);
        }
        cout << endl;

        filename = global.outmatrix_path + global.target + global.outmatrix_separ +
            global.source + global.outmatrix_ext;
        cout << "Writing gene matrix " << filename << " ..." << setw(9) << " ";
        if (!WriteMatrix(global.tosp, global.fromsp)) 
        {
            cerr << "Matrix output error: " << filename << endl;
            exit(7);
        }
        cout << endl;

        filename = global.outmatrix_path + global.source + global.outmatrix_separ +
            global.source + global.outmatrix_ext;
        cout << "Writing gene matrix " << filename << " ..." << setw(9) << " ";
        if (!WriteMatrix(global.fromsp, global.fromsp)) 
        {
            cerr << "Matrix output error: " << filename << endl;
            exit(7);
        }
        cout << endl;

        filename = global.outmatrix_path + global.target + global.outmatrix_separ +
            global.target + global.outmatrix_ext;
        cout << "Writing gene matrix " << filename << " ..." << setw(9) << " ";
        if (!WriteMatrix(global.tosp, global.tosp)) 
        {
            cerr << "Matrix output error: " << filename << endl;
            exit(7);
        }
        cout << endl;
    }

    if (global.quit == 0) exit(0);

    // 1st pass over genes of 'from' species to count homologs
    cout << "\nProcessing " << global.source << " genes ..." << endl;

    Species& to = global.Animals[global.tosp];
    int totalfr  = 0;   // total genes of 'from'
    int goodfr   = 0;   // 'from' genes with good E
    int genesfr  = 0;   // 'from' genes with strong arcs to paralogs 
    int arcsfr   = 0;   // number of strong arcs to paralogs in 'from'
    int genesto  = 0;   // 'from' genes with strong arcs to orthologs in 'to'
    int arcsto   = 0;   // number of strong arcs to orthologs in 'to'

    Species& from = global.Animals[global.fromsp];
    Giter git = from.genes.begin();
    for ( ; git != from.genes.end(); git++) 
    {
        if (global.genestep != 0 && totalfr % global.genestep == 0) 
        {
            cout << setw(9) << totalfr << " \r";
        }
        Gene& fgene = git->second;  // from gene (of new species)
        totalfr++;
        int self = -1;

        vector <Homolog> & vhff = fgene.homologs[global.fromsp];    // homologs in 'from'
        if (vhff.size() != 0)
        {
            goodfr++;
            int iself, i;
            for (i = 0; i < (int)vhff.size(); i++)      // find self
            {
                if (vhff[i].evalue > global.Elimit*lower_E)
                    continue;

                if (vhff[i].pgene == &fgene) 
                {
                    self = vhff[i].score;
                    iself = i;
                    break;
                }
            }
            if (self < 0)
            {
                cerr << "No self record for gene " << fgene.name << endl;
                //exit(8);
            }

            bool arcToPara = false;
            for (i = 0; i < (int)vhff.size(); i++)      // check weight
            {
                if (i == iself || vhff[i].evalue > global.Elimit*lower_E) 
                    continue;

                Gene& pgene = *(vhff[i].pgene);             // putative paralog             
                int self2 = -1;
                vector <Homolog> & vhtt = pgene.homologs[global.fromsp];
                for (int j = 0; j < (int)vhtt.size(); j++)  // find self2
                {
                    if (vhtt[j].evalue > global.Elimit*lower_E)
                        continue;

                    if (vhtt[j].pgene == &pgene)
                    {
                        self2 = vhtt[j].score;
                        break;
                    }
                }
                if (self2 < 0)
                {
                    cerr << "No self record for gene " << pgene.name << endl;
                    // exit(8);
                }

                if (2.0*vhff[i].score / (self + self2) >= global.pweight)
                {
                    arcsfr++;
                    arcToPara = true;
                }
                else        // disable paralogy
                {
                    vhff[i].evalue = big_E;
                    for (int j = 0; j < (int)vhtt.size(); j++)
                    {
                        if (vhtt[j].pgene == &fgene)
                        {
                            vhtt[j].evalue = big_E;
                            break;
                        }
                    }
                }
            }
            if (arcToPara)
                genesfr++;
        }

        vector <Homolog> & vhft = fgene.homologs[global.tosp];      // homologs in 'to'
        if (vhft.size() != 0) 
        {
            bool arcToOrtho = false;
            for (int i = 0; i < (int)vhft.size(); i++)  // check weight
            {
                Gene& tgene = *(vhft[i].pgene);             // putative ortholog
                vector <Homolog> & vhtt = tgene.homologs[global.nAnimals];  // self-scores
                int jself, self2 = -1;
                for (int j = 0; j < (int)vhtt.size(); j++)      // find self
                {
                    if (vhtt[j].evalue > global.Elimit*lower_E)
                        continue;

                    if (vhtt[j].pgene == &tgene)
                    {
                        self2 = vhtt[j].score;
                        jself = j;
                        break;
                    }
                }
                if (self2 < 0)
                {
                    cerr << "No self record for gene " << tgene.name << endl;
                    // exit(8);
                }

                if (2.0*vhft[i].score / (self + self2) >= global.oweight)
                {
                    arcsto++;
                    arcToOrtho = true;
                }
                else        // disable orthology
                {
                    vhft[i].evalue = big_E;
                    vector <Homolog> & vhtf = tgene.homologs[global.fromsp];
                    for (int j = 0; j < (int)vhtf.size(); j++)  
                    {
                        if (vhtf[j].pgene == &fgene)
                        {
                            vhtf[j].evalue = big_E;
                            break;
                        }
                    }
                }
            }
            if (arcToOrtho)
                genesto++;
        }
    }
    cout << "Genes of " << global.source << ":" << totalfr << "  Good:" 
        << goodfr << "  Have_para:" << genesfr << " [" << arcsfr 
        << "]  Have_ortho:" << genesto << " [" << arcsto << "]" << endl;

    if (global.quit == 1) exit(0);

    // Write table of paralogs with transitive closure of paralogy
    filename = global.paralogs_path + global.source + global.paralogs_ext;
    cout << "\nWriting paralogs to " << filename << " ..." << setw(9) << " ";
    ofstream pf;
    pf.open(filename.c_str());
    if (!pf || pf.bad()) 
    {
        cout << "Can't open " << filename << endl;
        exit(10);
    }

    int pargenes = 0;           // source genes having paralogs
    int parpairs = 0;           // number of paralogous pairs
    int geneno = 0;             // serial number of current gene
    int revert = 0;             // sub-weight pairs to revert
    int insert = 0;             // sub-threshold pairs to insert
    git = from.genes.begin();
    for ( ; git != from.genes.end(); git++, geneno++) {
        Gene& fgene = git->second;      // current gene to consider
        vector <PGene> paralogs;
        paralogs.push_back(&fgene);
        bool added = true;
        int lastpara = 0;               // # of first added paralog

        while (added)
        {               // collect paralogs
            added = false;
            bool done = false;
            // insert all paralogs not added yet
            vector <Homolog> & vhff = paralogs[lastpara]->homologs[global.fromsp];
            for (int i = 0; i < (int)vhff.size(); i++) 
            {
                if (vhff[i].evalue > global.Elimit*lower_E) 
                    continue;
                PGene pgene = vhff[i].pgene;
                if (LessID(pgene, &fgene))
                {
                    done = true;
                    break;
                }
                int k;
                for (k = 0; k < (int)paralogs.size(); k++)
                    if (paralogs[k] == pgene) break;
                if (k < (int)paralogs.size())
                    continue;   // already presents--check next paralog

                paralogs.push_back(pgene);
                added = true;
            }
            if (done)
            {
                paralogs.clear();
                break;
            }
            lastpara++;
            if (lastpara >= (int)paralogs.size()) break;
        }

        if (paralogs.size() <= 1) continue;     // non-paralogous gene encountered
        pargenes += (int)paralogs.size();

        sort(paralogs.begin(), paralogs.end(), LessID);

        for (int i = 0; i < (int)paralogs.size(); i++)
        {
            for (int j = i+1; j < (int)paralogs.size(); j++)
            {
                pf << paralogs[i]->name << "\t" << paralogs[j]->name << endl;
                pf << paralogs[j]->name << "\t" << paralogs[i]->name << endl;
                parpairs += 2;

                // add the two pairs to internal tables changing E-value if needed
                vector <Homolog> & vhij = paralogs[i]->homologs[global.fromsp];
                int k;
                for (k = 0; k < (int)vhij.size(); k++)
                {
                    if (vhij[k].pgene = paralogs[j])
                    {
                        if (vhij[k].evalue > global.Elimit) 
                        {
                            vhij[k].evalue = global.Elimit;     // revert arc
                            revert++;
                            break;
                        }
                    }
                }
                if (k >= (int)vhij.size())
                {
                    Homolog h(paralogs[j], global.Elimit, dummy_score);
                    vhij.push_back(h);                          // insert arc
                    insert++;
                }

                vector <Homolog> & vhji = paralogs[j]->homologs[global.fromsp];
                for (k = 0; k < (int)vhji.size(); k++)
                {
                    if (vhji[k].pgene = paralogs[i])
                    {
                        if (vhji[k].evalue > global.Elimit) 
                        {
                            vhji[k].evalue = global.Elimit;     // revert arc
                            revert++;
                            break;
                        }
                    }
                }
                if (k >= (int)vhji.size())
                {
                    Homolog h(paralogs[i], global.Elimit, dummy_score);
                    vhji.push_back(h);                          // insert arc
                    insert++;
                }
            }
        }
        if (outgenestep != 0 && geneno % outgenestep == 0)
            cout << back9 << setw(9) << geneno;
    }
    pf.close();
    cout << back9 << setw(9) << pargenes << " [" << parpairs << "]" << endl;
    cout << "Paralogous pairs reverted: " << revert << ", inserted: " << insert << endl;
    cout << "You should sort this file before use!" << endl;

    if (global.quit == 2) exit(0);

    // prepare for appending table of orthologs
    int ortgenes = 0;           // target genes with orthologs in source
    int ortpairs = 0;           // number of orthologous pairs
    int or1genes = 0;           // target genes with orthologs ONLY in source
    int or1pairs = 0;           // number of such orthologous pairs
    fstream df;
    if (global.addition_file.empty()) 
    {
        filename = global.homology_path + global.target + global.homology_ext;
        df.open(filename.c_str(), ios_base::out | ios_base::app);
    }
    else 
    {
        filename = global.homology_path + global.addition_file;
        df.open(filename.c_str(), ios_base::out);
    }
    if (!df || df.bad()) 
    {
        cout << "Can't write to " << filename << endl;
        exit(11);
    }
    cout << "\nWriting orthologs to " << filename << " ..." << setw(9) << " ";
    df << global.target << " " << global.myfields[field_gene] << "\t"
       << global.source << " " << global.myfields[field_gene] << endl;

    // 2nd pass over genes of 'to' species to process orthologs
    git = to.genes.begin();
    geneno = 0;
    for ( ; git != to.genes.end(); git++, geneno++) 
    {
        Gene& tgene = git->second;
        // check if tgene has orthologs in other species
        bool hasother = false;
        bool hasnew   = false;
        for (int i = 0; i < global.nAnimals; i++)
        {
            if (i == global.fromsp) continue;
            if ( !tgene.homologs[i].empty())
            {
                hasother = true;
                break;
            }
        }

        vector <Homolog> & vhtf = tgene.homologs[global.fromsp];
        for (int i = 0; i < (int)vhtf.size(); i++)
        {
            if (vhtf[i].evalue > global.Elimit) continue;
            hasnew = true;
            df << tgene.name << "\t" << vhtf[i].pgene->name << endl;
            if (hasother) ortpairs++;
            else          or1pairs++;
        }
        if (hasnew) 
        {
            if (hasother) ortgenes++;
            else          or1genes++;
        }
        if (outgenestep != 0 && geneno % outgenestep == 0)
            cout << back9 << setw(9) << geneno;
    }
    df.close();
    cout << back9 << setw(9) << ortgenes+or1genes << " [" << ortpairs+or1pairs << "]" << endl;

    cout << "Orthologs in " << global.target << " modified: " << ortgenes 
        << " [" << ortpairs << "], added: " << or1genes << " [" 
        << or1pairs << "]" << endl;
    cout << "You should sort this file before use!" << endl;

    // Analyze consistency of the 'from' species paralogs
    int f0  = 0;        // genes without orthologs in 'to'
    int f1  = 0;        // genes with 1 ortholog in 'to'
    int f2  = 0;        // genes with 2 orthologs in 'to'
    int f3  = 0;        // genes with many orthologs
    int p   = 0;        // total paralogous pairs
    int p2p = 0;        // paralogs to paralogs
    int p2s = 0;        // paralogs to same gene
    int p2n = 0;        // paralogs to non-paralogs
    int p1  = 0;        // one of paralogs to a gene
    int p0  = 0;        // none of paralogs to a gene
    int n   = 0;        // total non-paralogous pairs
    int n2p = 0;        // non-paralogs to paralogs
    int n2s = 0;        // non-paralogs to same gene
    int n2n = 0;        // non-paralogs to non-paralogs;
    int n1  = 0;        // one of non-paralogs to a gene
    int n0  = 0;        // none of non-paralogs to a gene
    bool topara, tosame;

    git = from.genes.begin();
    for ( ; git != from.genes.end(); git++)
    {
        Gene& gi = git->second;
        vector <Homolog> & vhh = gi.homologs[global.fromsp];
        vector <Homolog> & vhi = gi.homologs[global.tosp];

        int ti = 0;      // number of orthologs in 'to'
        for (int k = 0; k < (int)vhi.size(); k++)
            if (vhi[k].evalue <= global.Elimit) ti++;
        switch (ti)
        {
            case 0: f0++; break;
            case 1: f1++; break;
            case 2: f2++; break;
            default: f3++;
        }

        Giter gjt = git;
        for (++gjt; gjt != from.genes.end(); gjt++)
        {
            Gene& gj = gjt->second;
            vector <Homolog> & vhj = gj.homologs[global.tosp];

            int tj = 0;      // number of orthologs in 'to'
            for (int k = 0; k < (int)vhj.size(); k++)
                if (vhj[k].evalue <= global.Elimit) tj++;

            // check paralogy of gi and gj
            bool arepara = false;
            for (int k = 0; k < (int)vhh.size(); k++)
            {
                if (vhh[k].evalue <= global.Elimit && vhh[k].pgene == &gj)
                {
                    arepara = true;
                    break;
                }
            }
            if (arepara)    p++;
            else            n++;

            // count variants
            if (ti == 0 && tj == 0)
            {
                if (arepara)    p0++;
                else            n0++;
            }
            else if (ti == 0 || tj == 0)
            {
                if (arepara)    p1++;
                else            n1++;
            }
            else
            {
                topara = false; 
                tosame = false; 
                for (int i1 = 0; i1 < (int)vhi.size(); i1++)
                {
                    if (vhi[i1].evalue > global.Elimit) continue;
                    for (int j1 = 0; j1 < (int)vhj.size(); j1++)
                    {
                        if (vhj[j1].evalue > global.Elimit) continue;

                        // check coinciding orthologs
                        if (vhi[i1].pgene == vhj[j1].pgene)
                        {
                            tosame = true;
                            continue;
                        }

                        // check paralogy of orthologs
                        vector <Homolog> & vhij = vhi[i1].pgene->homologs[global.tosp];
                        for (int k = 0; k < (int)vhij.size(); k++)
                        {
                            if (vhij[k].evalue <= global.Elimit && vhij[k].pgene == vhj[j1].pgene)
                            {
                                topara = true;
                                break;
                            }
                        }
                    }
                }
                if (topara)
                {
                    if (arepara)    p2p++;
                    else            n2p++;
                }
                else if (tosame)
                {
                    if (arepara)    p2s++;
                    else            n2s++;
                }
                else
                {
                    if (arepara)    p2n++;
                    else            n2n++;
                }
            }
        }
    }
    cout << "\n" << global.source << " genes total:" << from.genes.size()
        << "  0-orth:" << f0 << "  1-orth:" << f1 << "  2-orth:" << f2
        << "  many-orth:" << f3 << endl;
    cout << "Paralogous pairs total:" << p << "  p2p:" << p2p << "  p2n:" << p2n 
        << "  p2s:" << p2s << "  p1:" << p1 << "  p0:" << p0 << endl;
    cout << "Non-paralogous pairs total:" << n << "  n2p:" << n2p << "  n2n:" 
        << n2n << "  n2s:" << n2s << "  n1:" << n1 << "  n0:" << n0 << endl;

    finish = clock();
    double elapsed = (int)((double)(finish - start) / CLOCKS_PER_SEC);
    if (elapsed > 300) 
    {
        int minutes = (int)(elapsed / 60.0);
        cout << "\n" << minutes << "m";
    }
    else 
    {
        int seconds = (int)elapsed;
        cout << "\n" << seconds << "s";
    }
    cout << ": Completed OK." << endl;
}